import torch
from torch.autograd        import grad
from sparse_distance_bmp   import sparse_distance_bmp
from pykeops.torch         import Kernel

import numpy as np
from matplotlib import pyplot as plt
from scipy import misc
from time  import time

use_cuda = torch.cuda.is_available()
tensor   = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor 

plt.ion()
plt.show()

s2v = lambda x : tensor([x])


from scipy.ndimage.filters import gaussian_filter

def LoadImage(fname) :
    img = misc.imread(fname, flatten = True) # Grayscale
    img = gaussian_filter(img, 1, mode='nearest')
    img = (img[::-1, :])  / 255.
    img = np.swapaxes(img, 0,1 )
    return tensor( 1 - img )

# Load the pngs
dataset = "bars"
datasets = {
    "bars"     : ("data/hor_bars_1.png", "data/hor_bars_2.png"),
    "worms_c"  : ("data/worm_c.png",     "data/worm_d.png"),
}
if dataset == "knees_2D" :
    sample = 8
    smooth = True # As the structure is very thin, sampling on a grid every 8 pixels looks weird...
                  # we smooth the gradient a little bit to prevent this visualization artifact from being visible
    scale_grad = 0.04
else :
    sample = 4
    smooth = False
    scale_grad = 0.02

source = LoadImage(datasets[dataset][0])
target = LoadImage(datasets[dataset][1])

# The images are rescaled to fit into the unit square ==========================================
scale = source.shape[0]
affine = tensor( [[ 1, 0, 0 ],
                  [ 0, 1, 0 ] ]) / scale

# Parameters of our data attachment term =======================================================
experiments = {}


if True : # Warmup PyTorch
    experiments["warmup"] = {
        "formula"     : "kernel",
        "id"          : Kernel( "-distance(x,y)" ),
        "gamma"       : s2v( 1. ),
    }

if True : # Display dataset
    experiments["dataset"] = {
        "formula"     : "kernel",
        "id"          : Kernel( "-distance(x,y)" ),
        "gamma"       : s2v( 1. ),
    }

if True : # Energy distance
    experiments["energy_distance"] = {
        "formula"     : "kernel",
        "id"          : Kernel( "-distance(x,y)" ),
        "gamma"       : s2v( 1. ),
    }

if True : # Kernel
    experiments["gaussian_01"] = {
        "formula"     : "kernel",
        "id"          : Kernel( "gaussian(x,y)" ),
        "gamma"       : s2v( 1/ .1**2 ),
    }
    experiments["laplacian_01"] = {
        "formula"     : "kernel",
        "id"          : Kernel( "laplacian(x,y)" ),
        "gamma"       : s2v( 1/ .1**2 ),
    }

if False : # hausdorff_KL
    for sigma in [1., .1, .01] :
        experiments["hausdorff_KL_{:03.2f}".format(sigma)] = {
            "formula"     : "hausdorff_KL",
            "cost"        : "KL",
            "id"          : Kernel( "laplacian(x,y)" ),
            "gamma"       : s2v( 1./sigma**2 ),
            "scale"       : s2v(    sigma    ),
        }

if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L1_small_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   .01   ),
            "gamma"       : s2v( 1/.01**2),
        }

if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L1_medium_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   .05   ),
            "gamma"       : s2v( 1/.05**2),
        }

if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L1_medium2_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   .1   ),
            "gamma"       : s2v( 1/.1**2),
        }

if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L1_large_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   .5   ),
            "gamma"       : s2v( 1/.5**2),
        }

if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L1_inf_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   100.   ),
            "gamma"       : s2v( 1/100.**2),
        }

if False : # hausdorff_sinkhorn_KL
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_KL_L1_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "KL",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "laplacian(x,y)" ),
            "epsilon"     : s2v(   .1   ),
            "gamma"       : s2v( 1/.1**2),
        }

if True : # sinkhorn
    for nits in [1, 5, 10, 30, 100] :
        experiments["sinkhorn_L1_{}".format(nits)] = {
        "formula"     : "sinkhorn",
        "epsilon"     : s2v( .01 ),
        "tol"         : 1e-6,
        "nits"        : nits,
        "rho"         : -1., # Balanced transport. Make sure that both measures have equal mass!
        "transport_plan" : "heatmaps", #minimal_symmetric+heatmaps",
    }

if True : # sinkhorn
    for nits in [1, 5, 10, 30, 100] :
        experiments["sinkhorn_L2_{}".format(nits)] = {
        "formula"     : "sinkhorn",
        "epsilon"     : s2v( .01 ),
        "kernel"      : {
            "id"     : Kernel("gaussian(x,y)") ,
            "gamma"  :  1 / s2v( .01 )      },
        "tol"         : 1e-6,
        "nits"        : nits,
        "rho"         : -1., # Balanced transport. Make sure that both measures have equal mass!
        "transport_plan" : "heatmaps", #minimal_symmetric+heatmaps",
    }


if True : # hausdorff_sinkhorn
    for nits in [1, 2, 3, 5, 10] :
        experiments["hausdorff_sinkhorn_L2_medium_{}".format(nits)] = {
            "formula"        : "hausdorff",
            "cost"           : "bregman",
            "distance_field" : "sinkhorn",
            "nits"           : nits,
            "id"          : Kernel( "gaussian(x,y)" ),
            "epsilon"     : s2v(   .01   ),
            "gamma"       : s2v( 1/.01),
        }

# We'll save the output wrt. the number of iterations
display = True
plt.figure(figsize=(10,10))


def test(name, params, verbose=True) :
    params["kernel_heatmap_range"] = (0,1,100)

    # Compute the cost and gradient ============================================================
    t_0 = time()
    cost, grad_src, heatmaps = sparse_distance_bmp(params, source, target, 
                                                           affine, affine, 
                                                           normalize=True, info=display )
    t_1 = time()
    if verbose : print("{} : {:.2f}s, cost = {:.6f}".format( name, t_1-t_0, cost.item()) )

    # Display ==================================================================================
    grad_src = - grad_src # N.B.: We want to visualize a descent direction, not the opposite!
    if display :
        plt.clf()

        # Source + Target :
        source_plot = .3*source.cpu().numpy()
        target_plot = .3*target.cpu().numpy()
        img_plot    = np.dstack( ( np.ones(source_plot.shape) - target_plot, 
                                   np.ones(source_plot.shape) - source_plot - target_plot, 
                                   np.ones(source_plot.shape) - source_plot ) )
        
        plt.imshow( np.swapaxes(img_plot,0,1), origin="lower", extent=(0,1,0,1))

        if name != "dataset" :
            # Subsample the gradient field :
            grad_plot = grad_src.cpu().numpy()
            grad_plot = grad_plot[::sample, ::sample, :]
            if smooth :
                grad_plot = gaussian_filter(grad_plot, [1,1,0], mode='nearest')

            X,Y   = np.meshgrid( np.linspace(0, 1, grad_plot.shape[0]+1)[:-1] + .5/(sample*grad_plot.shape[0]), 
                                np.linspace(0, 1, grad_plot.shape[1]+1)[:-1] + .5/(sample*grad_plot.shape[1]) )
            U, V  = grad_plot[:,:,0], grad_plot[:,:,1]
            norms = np.sqrt( (grad_plot[:,:,1]**2 + grad_plot[:,:,0]**2) )
            mask  = (norms>0)
            scale = np.mean( norms[mask] )
            if smooth : mask  = (source_plot[::sample, ::sample]>0)

            Y, X, U, V = Y[mask], X[mask], U[mask], V[mask]
            plt.quiver( Y, X, U, V, 
                        scale = scale_grad*scale, scale_units="dots", color="#5CBF3A", zorder=3, width=0.0025)

            # "Distance" fields :
            if heatmaps is not None :
                heatmaps.plot(plt.gca(), {})
            
        # Save result in the "output/" folder :
        plt.savefig("output/2D/" + dataset + "_" + name +".png") 
        plt.pause(.01)


#test("hausdorff_KL_1_1", experiments["hausdorff_KL_1_1"], verbose=False) # one run for nothing
for name, params in experiments.items() :
    test(name, params)

print("Done.")
plt.show(block=True)
